#! Basic necessities to get the CLI running
import argparse
import os
import sys

# ! The actual download stuff
from spotdl.download.downloader import DownloadManager
from spotdl.search.songObj import SongObj
from spotdl.search.spotifyClient import SpotifyClient
# ! Song Search from different start points
from spotdl.search.utils import (
    get_playlist_tracks,
    get_album_tracks,
    get_artist_tracks,
    search_for_song,
)
from spotdl.download import ffmpeg

# ! Usage is simple - call:
#   'python __main__.py <links, search terms, tracking files separated by spaces>
# ! Eg.
# !      python __main__.py
# !          https://open.spotify.com/playlist/37i9dQZF1DWXhcuQw7KIeM?si=xubKHEBESM27RqGkqoXzgQ
# !          'old gods of asgard Control'
# !          https://open.spotify.com/album/2YMWspDGtbDgYULXvVQFM6?si=gF5dOQm8QUSo-NdZVsFjAQ
# !          https://open.spotify.com/track/08mG3Y1vljYA6bvDt4Wqkj?si=SxezdxmlTx-CaVoucHmrUA
# !
# ! Well, yeah its a pretty long example but, in theory, it should work like a charm.
# !
# ! A '.spotdlTrackingFile' is automatically  created with the name of the first song in the
# ! playlist/album or the name of the song supplied. We don't really re re re-query YTM and Spotify
# ! as all relevant details are stored to disk.
# !
# ! Files are cleaned up on download failure.
# !
# ! All songs are normalized to standard base volume. the soft ones are made louder,
# ! the loud ones, softer.
# !
# ! The progress bar is synched across multiple-processes (4 processes as of now), getting the
# ! progress bar to synch was an absolute pain, each process knows how much 'it' progressed,
# ! but the display has to be for the overall progress so, yeah... that took time.
# !
# ! spotdl will show you its true speed on longer download's - so make sure you try
# ! downloading a playlist.
# !
# ! still yet to try and package this but, in theory, there should be no errors.
# !
# !                                                          - cheerio! (Michael)
# !
# ! P.S. Tell me what you think. Up to your expectations?

# ! Script Help
help_notice = '''
To download a song run,
    spotdl [trackUrl]
    ex. spotdl https://open.spotify.com/track/0VjIjW4GlUZAMYd2vXMi3b?si=1stnMF5GSdClnIEARnJiiQ

To download a album run,
    spotdl [albumUrl]
    ex. spotdl https://open.spotify.com/album/4yP0hdKOZPNshxUOjY0cZj?si=AssgQQrVTJqptFe7X92jNg

To download a playlist, run:
    spotdl [playlistUrl]
    ex. spotdl https://open.spotify.com/playlist/37i9dQZF1E8UXBoz02kGID?si=oGd5ctlyQ0qblj_bL6WWow

To search for and download a song, run, with quotation marks:
Note: This is not accurate and often causes errors.
    spotdl [songQuery]
    ex. spotdl 'The Weeknd - Blinding Lights'

To resume a failed/incomplete download, run:
    spotdl [pathToTrackingFile]
    ex. spotdl 'The Weeknd - Blinding Lights.spotdlTrackingFile'

    .spotdlTrackingFiles are automatically created when a download starts and deleted on completion

You can queue up multiple download tasks by separating the arguments with spaces:
    spotdl [songQuery1] [albumUrl] [songQuery2] ... (order does not matter)
    ex. spotdl 'The Weeknd - Blinding Lights'
            https://open.spotify.com/playlist/37i9dQZF1E8UXBoz02kGID?si=oGd5ctlyQ0qblj_bL6WWow ...

spotDL downloads up to 4 songs in parallel, so for a faster experience,
download albums and playlist, rather than tracks.
'''


def console_entry_point():
    '''
    This is where all the console processing magic happens.
    Its super simple, rudimentary even but, it's dead simple & it works.
    '''
    arguments = parse_arguments()

    if ffmpeg.has_correct_version(arguments.ignore_ffmpeg_version) is False:
        sys.exit(1)

    SpotifyClient.init(
        client_id='5f573c9620494bae87890c0f08a60293',
        client_secret='212476d9b0f3472eaa762d90b19b0ba8'
    )

    if arguments.path:
        if not os.path.isdir(arguments.path):
            sys.exit("The output directory doesn't exist.")
        print(f"Will download to: {os.path.abspath(arguments.path)}")
        os.chdir(arguments.path)

    with DownloadManager() as downloader:

        for request in arguments.url:
            if 'open.spotify.com' in request and 'track' in request:
                print('Fetching Song...')
                song = SongObj.from_url(request)

                if song.get_youtube_link() is not None:
                    downloader.download_single_song(song)
                else:
                    print('Skipping %s (%s) as no match could be found on youtube' % (
                        song.get_song_name(), request
                    ))

            elif 'open.spotify.com' in request and 'album' in request:
                print('Fetching Album...')
                songObjList = get_album_tracks(request)

                downloader.download_multiple_songs(songObjList)

            elif 'open.spotify.com' in request and 'playlist' in request:
                print('Fetching Playlist...')
                songObjList = get_playlist_tracks(request)

                downloader.download_multiple_songs(songObjList)

            elif 'open.spotify.com' in request and 'artist' in request:
                print('Fetching artist...')
                artistObjList = get_artist_tracks(request)

                downloader.download_multiple_songs(artistObjList)

            elif request.endswith('.spotdlTrackingFile'):
                print('Preparing to resume download...')
                downloader.resume_download_from_tracking_file(request)

            else:
                print('Searching for song "%s"...' % request)
                try:
                    song = search_for_song(request)
                    downloader.download_single_song(song)
                except Exception as e:
                    print(e)


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog="spotdl",
        description=help_notice,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("url", type=str, nargs="+", help="URL to a song/album/playlist")
    parser.add_argument("-o", "--output", help="Output directory path", dest="path")
    parser.add_argument("--ignore-ffmpeg-version",
                        help="Ignore ffmpeg version", action="store_true")

    return parser.parse_args()


if __name__ == '__main__':
    console_entry_point()



# ===============
# === Imports ===
# ===============

from rich.console import Console
from rich.progress import BarColumn, TimeRemainingColumn, Progress, ProgressColumn
from rich.progress import Task
from rich.theme import Theme
from rich.style import StyleType
from rich.console import (
    JustifyMethod,
    detect_legacy_windows,
    OverflowMethod,
)
from rich.highlighter import Highlighter
from rich.text import Text

# ! These are not used, they're here for static type checking using mypy
from spotdl.search.songObj import SongObj
from typing import (
    List,
    Optional,
)
from pathlib import Path


# =============
# === Theme ===
# =============

custom_theme = Theme(
    {
        "bar.back": "grey23",
        "bar.complete": "rgb(165,66,129)",
        "bar.finished": "rgb(114,156,31)",
        "bar.pulse": "rgb(165,66,129)",
        "general": "green",
        "nonimportant": "rgb(40,100,40)",
        "progress.data.speed": "red",
        "progress.description": "none",
        "progress.download": "green",
        "progress.filesize": "green",
        "progress.filesize.total": "green",
        "progress.percentage": "green",
        "progress.remaining": "rgb(40,100,40)",
    }
)


# ========================================================
# === Modified rich Text Column to Support Fixed Width ===
# ========================================================


class SizedTextColumn(ProgressColumn):
    """A column containing text."""

    def __init__(
        self,
        text_format: str,
        style: StyleType = "none",
        justify: JustifyMethod = "left",
        markup: bool = True,
        highlighter: Highlighter = None,
        overflow: Optional[OverflowMethod] = None,
        width: int = 20,
    ) -> None:
        self.text_format = text_format
        self.justify = justify
        self.style = style
        self.markup = markup
        self.highlighter = highlighter
        self.overflow = overflow
        self.width = width
        super().__init__()

    def render(self, task: "Task") -> Text:
        _text = self.text_format.format(task=task)
        if self.markup:
            text = Text.from_markup(_text, style=self.style, justify=self.justify)
        else:
            text = Text(_text, style=self.style, justify=self.justify)
        if self.highlighter:
            self.highlighter.highlight(text)

        text.truncate(max_width=self.width, overflow=self.overflow, pad=True)
        return text


# =======================
# === Display Manager ===
# =======================


class DisplayManager:
    def __init__(self):

        # ! Change color system if "legacy" windows terminal to prevent wrong colors displaying
        self.isLegacy = detect_legacy_windows()

        # ! dumb_terminals automatically handled by rich. Color system is too but it is incorrect
        # ! for legacy windows ... so no color for y'all.
        self.console = Console(
            theme=custom_theme, color_system="truecolor" if not self.isLegacy else None
        )

        self._richProgressBar = Progress(
            SizedTextColumn(
                "[white]{task.description}",
                overflow="ellipsis",
                width=int(self.console.width / 3),
            ),
            SizedTextColumn("{task.fields[message]}", width=18, style="nonimportant"),
            BarColumn(bar_width=None, finished_style="green"),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeRemainingColumn(),
            console=self.console,
            # ! Normally when you exit the progress context manager (or call stop())
            # ! the last refreshed display remains in the terminal with the cursor on
            # ! the following line. You can also make the progress display disappear on
            # ! exit by setting transient=True on the Progress constructor
            transient=self.isLegacy,
        )

        self.songCount = 0
        self.overallTaskID = None
        self.overallProgress = 0
        self.overallTotal = 100
        self.overallCompletedTasks = 0
        self.quiet = False

        # ! Basically a wrapper for rich's: with ... as ...
        self._richProgressBar.__enter__()

    def print(self, *text, color="green"):
        """
        `text` : `any`  Text to be printed to screen
        Use this self.print to replace default print().
        """

        if self.quiet:
            return

        line = ""
        for item in text:
            line += str(item) + " "

        if color:
            self._richProgressBar.console.print(f"[{color}]{line}")
        else:
            self._richProgressBar.console.print(line)

    def set_song_count_to(self, songCount: int) -> None:
        """
        `int` `songCount` : number of songs being downloaded
        RETURNS `~`
        sets the size of the progressbar based on the number of songs in the current
        download set
        """

        # ! all calculations are based of the arbitrary choice that 1 song consists of
        # ! 100 steps/points/iterations
        self.songCount = songCount

        self.overallTotal = 100 * songCount

        if self.songCount > 4:
            self.overallTaskID = self._richProgressBar.add_task(
                description="Total",
                processID="0",
                message=f"{self.overallCompletedTasks}/{int(self.overallTotal / 100)} complete",
                total=self.overallTotal,
                visible=(not self.quiet),
            )

    def update_overall(self):
        """
        Updates the overall progress bar.
        """

        # If the overall progress bar exists
        if self.overallTaskID is not None:
            self._richProgressBar.update(
                self.overallTaskID,
                message=f"{self.overallCompletedTasks}/{int(self.overallTotal / 100)} complete",
                completed=self.overallProgress,
            )

    def new_progress_tracker(self, songObj):
        """
        returns new instance of `_ProgressTracker` that follows the `songObj` download subprocess
        """
        return _ProgressTracker(self, songObj)

    # def clear(self) -> None:
    #     '''
    #     clear the rich progress bar
    #     '''
    #     pass

    def close(self) -> None:
        """
        clean up rich
        """

        self._richProgressBar.stop()

    # def reset(self) -> None:
    #     '''
    #     restart progress for new download instance
    #     '''
    #     pass


# ========================
# === Progress Tracker ===
# ========================


class _ProgressTracker:
    def __init__(self, parent, songObj):
        self.parent = parent
        self.songObj = songObj

        self.progress = 0
        self.oldProgress = 0
        self.downloadID = 0
        self.status = ""

        self.taskID = self.parent._richProgressBar.add_task(
            description=songObj.get_display_name(),
            processID=str(self.downloadID),
            message="Download Started",
            total=100,
            completed=self.progress,
            start=False,
            visible=(not self.parent.quiet),
        )

    def notify_download_skip(self) -> None:
        """
        updates progress bar to reflect a song being skipped
        """

        self.progress = 100
        self.update("Skipping")

    def pytube_progress_hook(self, stream, chunk, bytes_remaining) -> None:
        """
        Progress hook built according to pytube's documentation. It is called each time
        bytes are read from youtube.
        """

        # ! This will be called until download is complete, i.e we get an overall

        fileSize = stream.filesize

        # ! How much of the file was downloaded this iteration scaled put of 90.
        # ! It's scaled to 90 because, the arbitrary division of each songs 100
        # ! iterations is (a) 90 for download (b) 5 for conversion & normalization
        # ! and (c) 5 for ID3 tag embedding
        iterFraction = len(chunk) / fileSize * 90

        self.progress = self.progress + iterFraction
        self.update("Downloading")

    def notify_youtube_download_completion(self) -> None:
        """
        updates progresbar to reflect a audio conversion being completed
        """

        self.progress = 90  # self.progress + 5
        self.update("Converting")

    def notify_conversion_completion(self) -> None:
        """
        updates progresbar to reflect a audio conversion being completed
        """

        self.progress = 95  # self.progress + 5
        self.update("Tagging")

    def notify_download_completion(self) -> None:
        """
        updates progresbar to reflect a download being completed
        """

        # ! Download completion implie ID# tag embedding was just finished
        self.progress = 100  # self.progress + 5
        self.update("Done")

    def notify_error(self, e, tb):
        """
        `e` : error message
        `tb` : traceback
        Freezes the progress bar and prints the traceback received
        """
        self.update(message="Error " + self.status)

        message = f"Error: {e}\tWhile {self.status}: {self.songObj.get_display_name()}\n {str(tb)}"
        self.parent.print(message, color="red")

    def update(self, message=""):
        """
        Called at every event.
        """

        self.status = message

        # The change in progress since last update
        delta = self.progress - self.oldProgress

        # Update the progress bar
        # ! `start_task` called everytime to ensure progress is remove from indeterminate state
        self.parent._richProgressBar.start_task(self.taskID)
        self.parent._richProgressBar.update(
            self.taskID,
            description=self.songObj.get_display_name(),
            processID=str(self.downloadID),
            message=message,
            completed=self.progress,
        )

        # If task is complete
        if self.progress == 100 or message == "Error":
            self.parent.overallCompletedTasks += 1
            if self.parent.isLegacy:
                self.parent._richProgressBar.remove_task(self.taskID)

        # Update the overall progress bar
        self.parent.overallProgress += delta
        self.parent.update_overall()

        self.oldProgress = self.progress


# ===============================
# === Download tracking class ===
# ===============================


class DownloadTracker:
    def __init__(self):
        self.songObjList = []
        self.saveFile: Optional[Path] = None

    def load_tracking_file(self, trackingFilePath: str) -> None:
        """
        `str` `trackingFilePath` : path to a .spotdlTrackingFile

        RETURNS `~`

        reads songsObj's from disk and prepares to track their download
        """

        # Attempt to read .spotdlTrackingFile, raise exception if file can't be read
        trackingFile = Path(trackingFilePath)
        if not trackingFile.is_file():
            raise FileNotFoundError(f"no such tracking file found: {trackingFilePath}")

        with trackingFile.open("rb") as file_handle:
            songDataDumps = eval(file_handle.read().decode())

        # Save path to .spotdlTrackingFile
        self.saveFile = trackingFile

        # convert song data dumps to songObj's
        # ! see, songObj.get_data_dump and songObj.from_dump for more details
        for dump in songDataDumps:
            self.songObjList.append(SongObj.from_dump(dump))

    def load_song_list(self, songObjList: List[SongObj]) -> None:
        """
        `list<songOjb>` `songObjList` : songObj's being downloaded

        RETURNS `~`

        prepares to track download of provided songObj's
        """

        self.songObjList = songObjList

        self.backup_to_disk()

    def get_song_list(self) -> List[SongObj]:
        """
        RETURNS `list<songObj>

        get songObj's representing songs yet to be downloaded
        """
        return self.songObjList

    def backup_to_disk(self):
        """
        RETURNS `~`

        backs up details of songObj's yet to be downloaded to a .spotdlTrackingFile
        """
        # remove tracking file if no songs left in queue
        # ! we use 'return None' as a convenient exit point
        if len(self.songObjList) == 0:
            if self.saveFile and self.saveFile.is_file():
                self.saveFile.unlink()
            return None

        # prepare datadumps of all songObj's yet to be downloaded
        songDataDumps = []

        for song in self.songObjList:
            songDataDumps.append(song.get_data_dump())

        # ! the default naming of a tracking file is $nameOfFirstSOng.spotdlTrackingFile,
        # ! it needs a little fixing because of disallowed characters in file naming
        if not self.saveFile:
            songName = self.songObjList[0].get_song_name()

            for disallowedChar in ["/", "?", "\\", "*", "|", "<", ">"]:
                if disallowedChar in songName:
                    songName = songName.replace(disallowedChar, "")

            songName = songName.replace('"', "'").replace(":", " - ")

            self.saveFile = Path(songName + ".spotdlTrackingFile")

        # backup to file
        # ! we use 'wb' instead of 'w' to accommodate your fav K-pop/J-pop/Viking music
        with open(self.saveFile, "wb") as file_handle:
            file_handle.write(str(songDataDumps).encode())

    def notify_download_completion(self, songObj: SongObj) -> None:
        """
        `songObj` `songObj` : songObj representing song that has been downloaded

        RETURNS `~`

        removes given songObj from download queue and updates .spotdlTrackingFile
        """

        if songObj in self.songObjList:
            self.songObjList.remove(songObj)

        self.backup_to_disk()

    def clear(self):
        self.songObjList = []
        self.saveFile = None



# ===============
# === Imports ===
# ===============

# ! the following are for the search provider to function
import typing
# ! Just for static typing
from typing import List

from rapidfuzz.fuzz import partial_ratio
from ytmusicapi import YTMusic
from bs4 import BeautifulSoup
from requests import get


# ================================
# === Note to readers / Coders ===
# ================================

# ! YTM search (the actual POST request), courtesy of Elliot G. (@rocketinventor)
# ! result parsing and song matching system by @Mikhail-Zex
# !
# ! Essentially, Without Elliot, you wouldn't have a YTM search provider at all.


# =======================
# === helper function ===
# =======================

def match_percentage(str1: str, str2: str, score_cutoff: float = 0) -> float:
    '''
    `str` `str1` : a random sentence

    `str` `str2` : another random sentence

    `float` `score_cutoff` : minimum score required to consider it a match
                             returns 0 when similarity < score_cutoff

    RETURNS `float`

    A wrapper around `rapidfuzz.fuzz.partial_ratio` to handle UTF-8 encoded
    emojis that usually cause errors
    '''

    # ! this will throw an error if either string contains a UTF-8 encoded emoji
    try:
        return partial_ratio(str1, str2, score_cutoff=score_cutoff)

    # ! we build new strings that contain only alphanumerical characters and spaces
    # ! and return the partial_ratio of that
    except:  # noqa:E722
        newStr1 = ''

        for eachLetter in str1:
            if eachLetter.isalnum() or eachLetter.isspace():
                newStr1 += eachLetter

        newStr2 = ''

        for eachLetter in str2:
            if eachLetter.isalnum() or eachLetter.isspace():
                newStr2 += eachLetter

        return partial_ratio(newStr1, newStr2, score_cutoff=score_cutoff)


# ========================================================================
# === Background functions/Variables (Not meant to be called directly) ===
# ========================================================================

# ! YTMusic api client
ytmApiClient = YTMusic()


def _parse_duration(duration: str) -> float:
    '''
    Convert string value of time (duration: "25:36:59") to a float value of seconds (92219.0)
    '''
    try:
        # {(1, "s"), (60, "m"), (3600, "h")}
        mappedIncrements = zip([1, 60, 3600], reversed(duration.split(":")))
        seconds = 0
        for multiplier, time in mappedIncrements:
            seconds += multiplier * int(time)
        return float(seconds)

    # ! This usually occurs when the wrong string is mistaken for the duration
    except (ValueError, TypeError, AttributeError):
        return 0.0


def _map_result_to_song_data(result: dict) -> dict:
    song_data = {}
    if result['resultType'] in ['song', 'video']:
        artists = ", ".join(map(lambda a: a['name'], result['artists']))
        video_id = result['videoId']
        if video_id is None:
            return {}
        song_data = {
            'name': result['title'],
            'type': result['resultType'],
            'artist': artists,
            'length': _parse_duration(result.get('duration', None)),
            'link': f'https://www.youtube.com/watch?v={video_id}',
            'position': 0
        }

        album = result.get('album')
        if album:
            song_data['album'] = album['name']

    return song_data


def _query_and_simplify(searchTerm: str) -> List[dict]:
    '''
    `str` `searchTerm` : the search term you would type into YTM's search bar

    RETURNS `list<dict>`

    For structure of dict, see comment at function declaration
    '''

    # ! For dict structure, see end of this function (~ln 268, ln 283) and chill, this
    # ! function ain't soo big, there are plenty of comments and blank lines

    # build and POST a query to YTM

    print(f'Searching for: {searchTerm}')
    searchResult = ytmApiClient.search(searchTerm)

    return list(map(_map_result_to_song_data, searchResult))


# =======================
# === Search Provider ===
# =======================

def search_and_order_ytm_results(songName: str, songArtists: List[str],
                                 songAlbumName: str, songDuration: int) -> dict:
    '''
    `str` `songName` : name of song

    `list<str>` `songArtists` : list containing name of contributing artists

    `str` `songAlbumName` : name of song's album

    `int` `songDuration`

    RETURNS `dict`

    each entry in the result if formated as {'$YouTubeLink': $matchValue, ...}; Match value
    indicates how good a match the result is the the given parameters. THe maximum value
    that $matchValue can take is 100, the least value is unbound.
    '''
    # Query YTM
    results = _query_and_simplify(get_ytm_search_query(songName, songArtists))

    # Assign an overall avg match value to each result
    linksWithMatchValue = {}

    for result in results:
        # ! skip results without videoId, this happens if you are country restricted or
        # ! video is unavailabe
        if result == {}:
            continue

        # ! If there are no common words b/w the spotify and YouTube Music name, the song
        # ! is a wrong match (Like Ruelle - Madness being matched to Ruelle - Monster, it
        # ! happens without this conditional)

        # ! most song results on youtube go by $artist - $songName, so if the spotify name
        # ! has a '-', this function would return True, a common '-' is hardly a 'common
        # ! word', so we get rid of it. Lower-caseing all the inputs is to get rid of the
        # ! troubles that arise from pythons handling of differently cased words, i.e.
        # ! 'Rhino' == 'rhino' is false though the word is same... so we lower-case both
        # ! sentences and replace any hypens(-)
        lowerSongName = songName.lower()
        lowerResultName = result['name'].lower()

        sentenceAWords = lowerSongName.replace('-', ' ').split(' ')

        commonWord = False

        # ! check for common word
        for word in sentenceAWords:
            if word != '' and word in lowerResultName:
                commonWord = True

        # ! if there are no common words, skip result
        if not commonWord:
            continue

        # Find artist match
        # ! match  = (no of artist names in result) / (no. of artist names on spotify) * 100
        artistMatchNumber = 0

        # ! we use fuzzy matching because YouTube spellings might be mucked up
        if result['type'] == 'song':
            for artist in songArtists:
                if match_percentage(artist.lower(), result['artist'].lower(), 85):
                    artistMatchNumber += 1
        else:
            # ! i.e if video
            for artist in songArtists:
                # ! something like match_percentage('rionos', 'aiobahn, rionos Motivation
                # ! (remix)' would return 100, so we're absolutely corrent in matching
                # ! artists to song name.
                if match_percentage(artist.lower(), result['name'].lower(), 85):
                    artistMatchNumber += 1

        # ! Skip if there are no artists in common, (else, results like 'Griffith Swank -
        # ! Madness' will be the top match for 'Ruelle - Madness')
        if artistMatchNumber == 0:
            continue

        artistMatch = (artistMatchNumber / len(songArtists)) * 100

        # Find name match
        nameMatch = round(match_percentage(result['name'], songName), ndigits=3)

        # Find album match
        # ! We assign an arbitrary value of 0 for album match in case of video results
        # ! from YouTube Music
        albumMatch = 0.0

        if result['type'] == 'song':
            album = result.get('album')
            if album:
                albumMatch = match_percentage(album, songAlbumName)

        # Find duration match
        # ! time match = 100 - (delta(duration)**2 / original duration * 100)
        # ! difference in song duration (delta) is usually of the magnitude of a few
        # ! seconds, we need to amplify the delta if it is to have any meaningful impact
        # ! wen we calculate the avg match value
        delta = result['length'] - songDuration
        nonMatchValue = (delta ** 2) / songDuration * 100

        timeMatch = 100 - nonMatchValue

        # the results along with the avg Match
        avgMatch = (artistMatch + albumMatch + nameMatch + timeMatch) / 4

        linksWithMatchValue[result['link']] = avgMatch

    return linksWithMatchValue


def get_ytm_search_query(songName: str, songArtists: List[str]) -> str:
    joined_artists = ', '.join(songArtists)
    return f'{joined_artists} - {songName}'


def search_and_get_best_match(songName: str, songArtists: List[str],
                              songAlbumName: str, songDuration: int) -> typing.Optional[str]:
    '''
    `str` `songName` : name of song

    `list<str>` `songArtists` : list containing name of contributing artists

    `str` `songAlbumName` : name of song's album

    `int` `songDuration` : duration of the song

    RETURNS `str` : link of the best match
    '''

    # ! This is lazy coding, sorry.
    results = search_and_order_ytm_results(
        songName, songArtists,
        songAlbumName, songDuration
    )

    if len(results) == 0:
        return None

    resultItems = list(results.items())
    sortedResults = sorted(resultItems, key=lambda x: x[1], reverse=True)

    # ! In theory, the first 'TUPLE' in sortedResults should have the highest match
    # ! value, we send back only the link
    return sortedResults[0][0]


def get_song_lyrics(song_name: str, song_artists: List[str]) -> str:
    """
    `str` `song_name` : name of song

    `list<str>` `song_artists` : list containing name of contributing artists

    RETURNS `str`: Lyrics of the song.

    Gets the metadata of the song.
    """

    headers = {
        'Authorization': 'Bearer alXXDbPZtK1m2RrZ8I4k2Hn8Ahsd0Gh_o076HYvcdlBvmc0ULL1H8Z8xRlew5qaG',
    }
    api_search_url = 'https://api.genius.com/search'
    search_query = f'{song_name} {", ".join(song_artists)}'

    api_response = get(
        api_search_url,
        params={'q': search_query},
        headers=headers
    ).json()

    song_id = api_response['response']['hits'][0]['result']['id']
    song_api_url = f'https://api.genius.com/songs/{song_id}'

    api_response = get(
        song_api_url,
        headers=headers
    ).json()

    song_url = api_response['response']['song']['url']

    genius_page = get(song_url)
    soup = BeautifulSoup(genius_page.text, 'html.parser')
    lyrics = soup.select_one('div.lyrics').get_text()

    return lyrics.strip()



import asyncio
import subprocess
import sys
import re


def has_correct_version(skip_version_check: bool = False) -> bool:
    process = subprocess.Popen(
        ['ffmpeg', '-version'],
        shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    proc_out, _ = process.communicate()

    if process.returncode == 127:
        print("FFmpeg was not found, spotDL cannot continue.", file=sys.stderr)
        return False

    if skip_version_check is False:
        result = re.search(r"ffmpeg version \w?(\d+\.)?(\d+)", proc_out.decode("utf-8"))

        if result is None:
            print("Your FFmpeg version couldn't be detected", file=sys.stderr)
            return False

        version = result.group(0).replace("ffmpeg version ", "")
        if float(version) < 4.3:
            print(f"Your FFmpeg installation is too old ({version}), please update to 4.3+\n",
                  file=sys.stderr)
            return False

    return True


async def convert(trackAudioStream, downloadedFilePath, convertedFilePath) -> bool:
    # convert downloaded file to MP3 with normalization

    # ! -af loudnorm=I=-7:LRA applies EBR 128 loudness normalization algorithm with
    # ! intergrated loudness target (I) set to -17, using values lower than -15
    # ! causes 'pumping' i.e. rhythmic variation in loudness that should not
    # ! exist -loud parts exaggerate, soft parts left alone.
    # !
    # ! dynaudnorm applies dynamic non-linear RMS based normalization, this is what
    # ! actually normalized the audio. The loudnorm filter just makes the apparent
    # ! loudness constant
    # !
    # ! apad=pad_dur=2 adds 2 seconds of silence toward the end of the track, this is
    # ! done because the loudnorm filter clips/cuts/deletes the last 1-2 seconds on
    # ! occasion especially if the song is EDM-like, so we add a few extra seconds to
    # ! combat that.
    # !
    # ! -acodec libmp3lame sets the encoded to 'libmp3lame' which is far better
    # ! than the default 'mp3_mf', '-abr true' automatically determines and passes the
    # ! audio encoding bitrate to the filters and encoder. This ensures that the
    # ! sampled length of songs matches the actual length (i.e. a 5 min song won't display
    # ! as 47 seconds long in your music player, yeah that was an issue earlier.)

    command = (
        'ffmpeg -v quiet -y -i "%s" -acodec libmp3lame -abr true '
        f"-b:a {trackAudioStream.bitrate} "
        '-af "apad=pad_dur=2, dynaudnorm, loudnorm=I=-17" "%s"'
    )

    # ! bash/ffmpeg on Unix systems need to have excape char (\) for special characters: \$
    # ! alternatively the quotes could be reversed (single <-> double) in the command then
    # ! the windows special characters needs escaping (^): ^\  ^&  ^|  ^>  ^<  ^^

    if sys.platform == "win32":
        formattedCommand = command % (
            str(downloadedFilePath),
            str(convertedFilePath),
        )
    else:
        formattedCommand = command % (
            str(downloadedFilePath).replace("$", r"\$"),
            str(convertedFilePath).replace("$", r"\$"),
        )

    process = await asyncio.subprocess.create_subprocess_shell(
        formattedCommand,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    _, proc_err = await process.communicate()

    if process.returncode != 0:
        message = (
            f"ffmpeg returned an error ({process.returncode})"
            f"\nthe ffmpeg command was \"{formattedCommand}\""
            "\nffmpeg gave this output:"
            "\n=====\n"
            f"{proc_err.decode('utf-8')}"
            "\n=====\n"
        )

        print(message, file=sys.stderr)
        return False

    return True



import json
import sys

import pytest

from spotdl.__main__ import console_entry_point, help_notice
from spotdl.download.downloader import DownloadManager
from spotdl.search.spotifyClient import SpotifyClient
from spotdl.download import ffmpeg

from tests.utils import tracking_files

ORIGINAL_INITIALIZE = SpotifyClient.init


def new_initialize(client_id, client_secret):
    """This function allows calling `initialize()` multiple times"""
    try:
        return SpotifyClient()
    except:
        return ORIGINAL_INITIALIZE(client_id, client_secret)


@pytest.fixture()
def patch_dependencies(mocker, monkeypatch):
    """This is a helper fixture to patch out everything that shouldn't be called here"""
    monkeypatch.setattr(SpotifyClient, "init", new_initialize)
    monkeypatch.setattr(DownloadManager, "__init__", lambda _: None)
    monkeypatch.setattr(ffmpeg, "has_correct_version", lambda *_: True)
    mocker.patch.object(DownloadManager, "download_single_song", autospec=True)
    mocker.patch.object(
        DownloadManager, "download_multiple_songs", autospec=True)
    mocker.patch.object(
        DownloadManager, "resume_download_from_tracking_file", autospec=True)
    mocker.patch.object(DownloadManager, "__exit__", autospec=True)


@pytest.mark.parametrize("argument", ["-h", "--help"])
def test_show_help(capsys, monkeypatch, argument):
    """The --help, -h switches or no arguments should display help message"""

    # `dummy` is an initial argument, which represents file path.
    # in real word sys.argv when no arguments are supplied contains just the script file path
    cli_args = ["dummy"]
    if argument:
        cli_args.append(argument)
    monkeypatch.setattr(sys, "argv", cli_args)

    with pytest.raises(SystemExit):
        console_entry_point()

    out, _ = capsys.readouterr()
    assert help_notice in out


@pytest.mark.vcr()
def test_download_a_single_song(capsys, patch_dependencies, monkeypatch):
    """First example from the help - using the track url should trigger song download."""
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "dummy",
            "https://open.spotify.com/track/08mG3Y1vljYA6bvDt4Wqkj?si=SxezdxmlTx-CaVoucHmrUA",
        ],
    )

    console_entry_point()

    out, err = capsys.readouterr()
    assert "Fetching Song...\n" in out

    assert DownloadManager.download_single_song.call_count == 1
    assert DownloadManager.download_multiple_songs.call_count == 0


@pytest.mark.vcr()
def test_download_an_album(capsys, patch_dependencies, monkeypatch):
    """Second example - download whole album."""
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "dummy",
            "https://open.spotify.com/album/2YMWspDGtbDgYULXvVQFM6?si=gF5dOQm8QUSo-NdZVsFjAQ",
        ],
    )

    console_entry_point()

    out, err = capsys.readouterr()
    assert "Fetching Album...\n" in out

    assert DownloadManager.download_multiple_songs.call_count == 1
    assert DownloadManager.download_single_song.call_count == 0


@pytest.mark.vcr()
def test_download_a_playlist(capsys, patch_dependencies, monkeypatch):
    """
    Third example - download a playlist.
    The playlist URL is different from the example, as the original one was to big.
    """
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "dummy",
            "https://open.spotify.com/playlist/0slbokxiWCo9egF9UhVmmI",
        ],
    )

    console_entry_point()

    out, err = capsys.readouterr()
    assert "Fetching Playlist...\n" in out

    assert DownloadManager.download_multiple_songs.call_count == 1
    assert DownloadManager.download_single_song.call_count == 0


@pytest.mark.vcr()
def test_search_and_download(capsys, patch_dependencies, monkeypatch):
    """Fourth example - search for a song, which cannot be found."""
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "dummy",
            "The HU - Sugaan Essenna",
        ],
    )

    # with pytest.raises(Exception, match="No song matches found on Spotify"):
    console_entry_point()

    out, err = capsys.readouterr()
    assert out == (
        'Searching for song "The HU - Sugaan Essenna"...\n'
        'No song matches found on Spotify\n')

    assert DownloadManager.download_multiple_songs.call_count == 0
    assert DownloadManager.download_single_song.call_count == 0


@pytest.mark.vcr()
def test_use_tracking_file(capsys, patch_dependencies, monkeypatch, fs):
    """Fifth example - use a spotdlTrackingFile."""
    fs.create_file("Back In Black.spotdlTrackingFile",
                   contents=json.dumps(tracking_files.back_in_black))
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "dummy",
            "Back In Black.spotdlTrackingFile",
        ],
    )

    console_entry_point()

    out, err = capsys.readouterr()
    assert out == 'Preparing to resume download...\n'

    assert DownloadManager.resume_download_from_tracking_file.call_count == 1
    assert DownloadManager.download_multiple_songs.call_count == 0
    assert DownloadManager.download_single_song.call_count == 0


@pytest.mark.vcr()
def test_multiple_elements(capsys, patch_dependencies, monkeypatch):
    """Last example - chaining tasks. Download two songs and search for one."""
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "dummy",
            "https://open.spotify.com/track/08mG3Y1vljYA6bvDt4Wqkj?si=SxezdxmlTx-CaVoucHmrUA",
            "https://open.spotify.com/track/2SiXAy7TuUkycRVbbWDEpo",
            "The HU - Sugaan Essenna",
        ],
    )

    console_entry_point()

    out, err = capsys.readouterr()
    assert 'Fetching Song...\n' in out
    assert 'Searching for song "The HU - Sugaan Essenna"...\n' in out
    assert 'No song matches found on Spotify\n' in out

    assert DownloadManager.download_single_song.call_count == 2
    assert DownloadManager.download_multiple_songs.call_count == 0



from spotipy import Spotify
from spotipy.oauth2 import SpotifyClientCredentials


class Singleton(type):
    """
    Singleton metaclass for SpotifyClient. Ensures that SpotifyClient is not
    instantiated without prior initialization. Every other instantiation of
    SpotifyClient will return the same instance.
    """
    _instance = None

    def __call__(cls):
        if cls._instance is None:
            raise Exception('Spotify client not created. Call SpotifyClient.init'
                            '(client_id, client_secret) first.')
        return cls._instance

    def init(cls, client_id: str, client_secret: str) -> "Singleton":
        '''
        `str` `client_id` : client id from your spotify account

        `str` `client_secret` : client secret for your client id

        creates and caches a spotify client if a client doesn't exist. Can only be called
        once, multiple calls will cause an Exception.
        '''

        # check if initialization has been completed, if yes, raise an Exception
        if cls._instance and cls._instance.is_initialized():
            raise Exception('A spotify client has already been initialized')

        credentialManager = SpotifyClientCredentials(
            client_id=client_id,
            client_secret=client_secret
        )
        cls._instance = super().__call__(client_credentials_manager=credentialManager)
        return cls._instance


class SpotifyClient(Spotify, metaclass=Singleton):
    """
    This is the Spotify client meant to be used in the app. Has to be initialized first by
    calling `SpotifyClient.init(client_id, client_secret)`.
    """
    _initialized = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initialized = True

    def is_intialized(self):
        return self._initialized



from typing import List

from spotdl.search.provider import search_and_get_best_match, get_song_lyrics
from spotdl.search.spotifyClient import SpotifyClient


class SongObj():
    # ! This can be accessed as songObj.searchProvider. songObj acts like a namespace
    # ! it allows us a convenient way of setting a search provider without using globals
    searchProvider = search_and_get_best_match

    # ====================
    # === Constructors ===
    # ====================
    def __init__(self, rawTrackMeta, rawAlbumMeta, rawArtistMeta, youtubeLink, lyrics):
        self.__rawTrackMeta = rawTrackMeta
        self.__rawAlbumMeta = rawArtistMeta
        self.__rawArtistMeta = rawArtistMeta
        self.__youtubeLink = youtubeLink
        self.__lyrics = lyrics

    # ! constructors here are a bit mucky, there are two different constructors for two
    # ! different use cases, hence the actual __init__ function does not exist

    # ! Note, since the following are class methods, an instance of songObj is initialized
    # ! and passed to them
    @classmethod
    def from_url(cls, spotifyURL: str):
        # check if URL is a playlist, user, artist or album, if yes raise an Exception,
        # else procede
        if not ('open.spotify.com' in spotifyURL and 'track' in spotifyURL):
            raise Exception('passed URL is not that of a track: %s' % spotifyURL)

        # query spotify for song, artist, album details
        spotifyClient = SpotifyClient()

        rawTrackMeta = spotifyClient.track(spotifyURL)

        primaryArtistId = rawTrackMeta['artists'][0]['id']
        rawArtistMeta = spotifyClient.artist(primaryArtistId)

        albumId = rawTrackMeta['album']['id']
        rawAlbumMeta = spotifyClient.album(albumId)

        # get best match from the given provider
        songName = rawTrackMeta['name']

        albumName = rawTrackMeta['album']['name']

        duration = round(
            rawTrackMeta['duration_ms'] / 1000,
            ndigits=3
        )

        contributingArtists = []

        for artist in rawTrackMeta['artists']:
            contributingArtists.append(artist['name'])

        youtubeLink = SongObj.searchProvider(
            songName,
            contributingArtists,
            albumName,
            duration
        )

        # try to get lyrics from Genius
        try:
            lyrics = get_song_lyrics(songName, contributingArtists)
        except (AttributeError, IndexError):
            lyrics = ""

        return cls(
            rawTrackMeta, rawAlbumMeta,
            rawArtistMeta, youtubeLink,
            lyrics
        )

    @classmethod
    def from_dump(cls, dataDump: dict):
        rawTrackMeta = dataDump['rawTrackMeta']
        rawAlbumMeta = dataDump['rawAlbumMeta']
        rawArtistMeta = dataDump['rawAlbumMeta']
        youtubeLink = dataDump['youtubeLink']
        lyrics = dataDump['lyrics']

        return cls(
            rawTrackMeta, rawAlbumMeta,
            rawArtistMeta, youtubeLink,
            lyrics
        )

    def __eq__(self, comparedSong) -> bool:
        if comparedSong.get_data_dump() == self.get_data_dump():
            return True
        else:
            return False

    # ================================
    # === Interface Implementation ===
    # ================================

    def get_youtube_link(self) -> str:
        return self.__youtubeLink

    # ! Song Details:

    # ! 1. Name
    def get_song_name(self) -> str:
        ''''
        returns songs's name.
        '''

        return self.__rawTrackMeta['name']

    # ! 2. Track Number
    def get_track_number(self) -> int:
        '''
        returns song's track number from album (as in weather its the first
        or second or third or fifth track in the album)
        '''

        return self.__rawTrackMeta['track_number']

    # ! 3. Genres
    def get_genres(self) -> List[str]:
        '''
        returns a list of possible genres for the given song, the first member
        of the list is the most likely genre. returns None if genre data could
        not be found.
        '''

        return self.__rawAlbumMeta['genres'] + self.__rawArtistMeta['genres']

    # ! 4. Duration
    def get_duration(self) -> float:
        '''
        returns duration of song in seconds.
        '''

        return round(self.__rawTrackMeta['duration_ms'] / 1000, ndigits=3)

    # ! 5. All involved artists
    def get_contributing_artists(self) -> List[str]:
        '''
        returns a list of all artists who worked on the song.
        The first member of the list is likely the main artist.
        '''

        # we get rid of artist name that are in the song title so
        # naming the song would be as easy as
        # $contributingArtists + songName.mp3, we would want to end up with
        # 'Jetta, Mastubs - I'd love to change the world (Mastubs remix).mp3'
        # as a song name, it's dumb.

        contributingArtists = []

        for artist in self.__rawTrackMeta['artists']:
            contributingArtists.append(artist['name'])

        return contributingArtists
    # ! 6. Disc Number
    def get_disc_number(self) -> int:
        return self.__rawTrackMeta['disc_number']

    # ! 6. Lyrics
    def get_lyrics(self):
        '''
        returns the lyrics of the song if found on Genius
        '''

        return self.__lyrics

    # ! 6. Display Name
    def get_display_name(self) -> str:
        ''''
        returns songs's display name.
        '''

        return str(str(", ".join(self.get_contributing_artists()) + " - " + self.get_song_name()))

    # ! Album Details:

    # ! 1. Name

    def get_album_name(self) -> str:
        '''
        returns name of the album that the song belongs to.
        '''

        return self.__rawTrackMeta['album']['name']

    # ! 2. All involved artist
    def get_album_artists(self) -> List[str]:
        '''
        returns list of all artists who worked on the album that
        the song belongs to. The first member of the list is likely the main
        artist.
        '''

        albumArtists = []

        for artist in self.__rawTrackMeta['album']['artists']:
            albumArtists.append(artist['name'])

        return albumArtists

    # ! 3. Release Year/Date
    def get_album_release(self) -> str:
        '''
        returns date/year of album release depending on what data is available.
        '''

        return self.__rawTrackMeta['album']['release_date']

    # ! Utilities for genuine use and also for metadata freaks:

    # ! 1. Album Art URL
    def get_album_cover_url(self) -> str:
        '''
        returns url of the biggest album art image available.
        '''

        return self.__rawTrackMeta['album']['images'][0]['url']

    # ! 2. All the details the spotify-api can provide
    def get_data_dump(self) -> dict:
        '''
        returns a dictionary containing the spotify-api responses as-is. The
        dictionary keys are as follows:
            - rawTrackMeta      spotify-api track details
            - rawAlbumMeta      spotify-api song's album details
            - rawArtistMeta     spotify-api song's artist details

        Avoid using this function, it is implemented here only for those super
        rare occasions where there is a need to look up other details. Why
        have to look it up seperately when it's already been looked up once?
        '''

        # ! internally the only reason this exists is that it helps in saving to disk

        return {
            'youtubeLink': self.__youtubeLink,
            'rawTrackMeta': self.__rawTrackMeta,
            'rawAlbumMeta': self.__rawAlbumMeta,
            'rawArtistMeta': self.__rawArtistMeta,
            'lyrics': self.__lyrics
        }



__all__ = [
    'provider',
    'songObj',
    'spotifyClient',
    'utils'
]

# ! You should be able to do all you want with just theese three lines
# ! from spotdl.search.spotifyClient    import initialize
# ! from spotdl.search.songObj          import songObj
# ! from spotdl.search.utils            import *
